# ruff: noqa: D100, D101, D103

import pytest

from xivapy.client import Client
from xivapy.model import Model


@pytest.fixture
def client():
    return Client()


@pytest.fixture
def sample_model_data():
    # TODO: get some real data
    return {
        'row_id': 1,
        'Name': 'Test Item',
        'Level': 50,
    }


class TestModel(Model):
    __sheetname__ = 'TestSheet'
    row_id: int
    name: str = ''
    level: int = 0


@pytest.fixture
def test_model_class():
    return TestModel



"""xivapy's version."""

from typing import Final
import importlib.util
from importlib.metadata import version

__all__ = ['VERSION']


def is_editable_install() -> bool:
    """Determines if the package is in development or not."""
    try:
        spec = importlib.util.find_spec('xivapy')
        if spec and spec.origin:
            return 'site-packages' not in spec.origin
    except Exception:
        # Assume dev?
        return True
    # Assume dev
    return True


VERSION: Final[str] = 'dev' if is_editable_install() else version('xivapy')



"""xivapy, an async Python client for XIVAPI for Final Fantasy XIV."""

from xivapy.client import Client, SearchResult
from xivapy.query import Query, QueryBuilder, Group
from xivapy.model import FieldMapping, Model

# TODO: maybe scope this so people can xivapi.types.Format?
# For now the api surface is small, so we don't have conflicts anyway
from xivapy.types import LangDict, Format
import xivapy.exceptions as exceptions

__all__ = [
    'Client',
    'SearchResult',
    'Query',
    'QueryBuilder',
    'Group',
    'FieldMapping',
    'Model',
    'LangDict',
    'Format',
    'exceptions',
]



"""xivapy Model-related classes."""

from typing import Optional, Any
from dataclasses import dataclass

from pydantic import BaseModel, model_validator

__all__ = [
    'Model',
    'FieldMapping',
]


@dataclass
class FieldMapping:
    """Map a single model field to multiple XIVAPI fields."""

    base_field: str
    languages: Optional[list[str]] = None
    raw: bool = False
    html: bool = False
    custom_spec: Optional[str] = None

    def to_field_specs(self) -> list[str]:
        """Transforms a Model field into an xivapi-understood field."""
        if self.custom_spec:
            return [self.custom_spec]

        specs = []
        if self.languages:
            for lang in self.languages:
                specs.append(f'{self.base_field}@lang({lang})')
        elif self.raw:
            specs.append(f'{self.base_field}@as(raw)')
        elif self.html:
            specs.append(f'{self.base_field}@as(html)')
        else:
            specs.append(self.base_field)

        return specs


class Model(BaseModel):
    """Base model for all xivapy queries."""

    __sheetname__: Optional[str] = None

    model_config = {'populate_by_name': True}

    @classmethod
    def get_sheet_name(cls) -> str:
        """Returns the sheet name, defaulting to the class name if __sheetname__ not set."""
        if cls.__sheetname__:
            return cls.__sheetname__
        return cls.__name__

    @classmethod
    def get_fields_str(cls) -> str:
        """Returns all model fields as a comma-separated string list for XIVAPI queries."""
        return ','.join(cls.get_xivapi_fields())

    @classmethod
    def _get_field_mapping(cls, field_info) -> Optional[FieldMapping]:
        """Gets the xivapy-specific metadata for a field, if one is defined."""
        if hasattr(field_info, 'metadata') and field_info.metadata:
            for metadata in field_info.metadata:
                if isinstance(metadata, FieldMapping):
                    return metadata
        return None

    @classmethod
    def get_xivapi_fields(cls) -> set[str]:
        """Get a set of all defined field names."""
        fields = set()

        for field_name, field_info in cls.model_fields.items():
            default_field = field_info.alias or field_name
            mapping = cls._get_field_mapping(field_info)

            if mapping:
                for spec in mapping.to_field_specs():
                    fields.add(spec)
            else:
                fields.add(default_field)

        return fields

    @classmethod
    def _process_mapped_field(
        cls, data: dict[str, Any], model_field: str, mapping: FieldMapping
    ) -> dict[str, Any]:
        if mapping.languages:
            # Collect lang variants
            lang_dict = {}
            for lang in mapping.languages:
                field_key = f'{mapping.base_field}@lang({lang})'
                if field_key in data:
                    lang_dict[lang] = data[field_key]
            if lang_dict:
                data[model_field] = lang_dict

        elif mapping.raw:
            field_key = f'{mapping.base_field}@as(raw)'
            if field_key in data:
                data[model_field] = data[field_key]

        elif mapping.html:
            field_key = f'{mapping.base_field}@as(html)'
            if field_key in data:
                data[model_field] = data[field_key]

        elif mapping.custom_spec:
            if mapping.custom_spec in data:
                data[model_field] = data[mapping.custom_spec]

        else:
            # Handle nested fields
            if '.' in mapping.base_field:
                value = cls._extract_nested_field(data, mapping.base_field)
                if value is not None:
                    data[model_field] = value
            elif mapping.base_field in data:
                data[model_field] = data[mapping.base_field]

        return data

    @classmethod
    def _extract_nested_field(cls, data: dict, field_path: str) -> Any:
        """Extract nested field data from xivapi response using dot notation (e.g., 'ContentType.Name')."""
        parts = field_path.split('.')
        current = data

        for i, part in enumerate(parts):
            if part in current:
                obj = current[part]

                # Navigate through the dark fields
                if isinstance(obj, dict):
                    if 'fields' in obj and len(parts) > i + 1:
                        current = obj['fields']
                    elif i == len(parts) - 1:
                        # we've gone to the bottom of the fields
                        return obj
                    else:
                        current = obj
                else:
                    return obj if i == len(parts) - 1 else None
            else:
                return None
        return current

    @model_validator(mode='before')
    @classmethod
    def process_xivapi_response(cls, data: dict[str, Any]) -> dict[str, Any]:
        """Model validator that processes xivapi-specific response data for pydantic validation."""
        if not isinstance(data, dict):
            return data

        for field_name, field_info in cls.model_fields.items():
            mapping = cls._get_field_mapping(field_info)
            if mapping:
                data = cls._process_mapped_field(data, field_name, mapping)

        return data



"""API responses to test against."""

VERSIONS_RESPONSE = {
    'versions': [
        {'names': ['7.3x1']},
        {'names': ['latest']},
    ]
}

SHEETS_RESPONSE = {
    'sheets': [
        {'name': 'Item'},
        {'name': 'ContentFinderCondition'},
        {'name': 'Quest'},
    ]
}

BASIC_SEARCH_RESPONSE = {
    'results': [
        {
            'score': 1.0,
            'sheet': 'TestSheet',
            'row_id': 1,
            'fields': {'Name': 'Test Item', 'Level': 50},
        }
    ],
    'next': None,
}

SEARCH_RESPONSE_PAGE_1 = {
    'results': [
        {
            'score': 1.0,
            'sheet': 'TestSheet',
            'row_id': 23,
            'fields': {'Name': 'Test Item', 'Level': 89},
        }
    ],
    'next': '28433b5b-7860-4395-88df-17c75c173a7c',
}
SEARCH_RESPONSE_PAGE_2 = {
    'results': [
        {
            'score': 0.899,
            'sheet': 'TestSheet',
            'row_id': 212,
            'fields': {'Name': 'Another Test Item', 'Level': 50},
        }
    ],
    'next': None,
}

SHEET_ROW_RESPONSE = {'row_id': 1, 'fields': {'Name': 'Test Item', 'Level': 50}}



"""Tests for query.py."""

import pytest

from xivapy.query import QueryBuilder
from xivapy.exceptions import QueryBuildError


def test_compound_where_query():
    """Test multiple items under .where()."""
    query = QueryBuilder().where(Name='Test', Level=50)
    assert query.build() == 'Name="Test" Level=50'


def test_contains_query():
    """Test contains query."""
    query = QueryBuilder().contains(Name='sword')
    assert query.build() == 'Name~"sword"'


def test_comparison_operators():
    """Test both > and < operations."""
    query = QueryBuilder().gt(Level=50).lt(Level=100)
    assert query.build() == 'Level>50 Level<100'


def test_required_modifier():
    """Test marking a query as required (+)."""
    query = QueryBuilder().where(Name='Test').required()
    assert query.build() == '+Name="Test"'


def test_excluded_modifier():
    """Test marking a query as excluded (-)."""
    query = QueryBuilder().where(Name='Test').excluded()
    assert query.build() == '-Name="Test"'


def test_empty_query():
    """Test an empty query."""
    query = QueryBuilder()
    assert query.build() == ''


def test_comparison_equal_operators():
    """Test >= and <= operators."""
    query = QueryBuilder().gte(Level=50).lte(Level=60)
    assert query.build() == 'Level>=50 Level<=60'


def test_or_any_with_query_builder():
    """Test query grouping using only QueryBuilder."""
    query = (
        QueryBuilder()
        .contains(Name='the')
        .or_any(
            QueryBuilder().contains(Name='extreme'),
            QueryBuilder().contains(Name='savage'),
        )
    )
    assert query.build() == 'Name~"the" (Name~"extreme" Name~"savage")'


def test_bool_values_serialize_correctly():
    """Test that bool values in a query correctly serialize to their lowercase forms."""
    query = QueryBuilder().where(One=True, Two=False).build()

    assert 'One=true' in query
    assert 'Two=false' in query


def test_required_exclude_on_same_query():
    """Test making a query as required and excluded."""
    with pytest.raises(QueryBuildError):
        query = (
            QueryBuilder()
            .where(Name='The Winding Spirals of Bahumhant')
            .required()
            .excluded()
        )
        query.build()



"""Types used in xivapy."""

from typing import Literal, TypedDict

__all__ = ['Format', 'LangDict']

Format = Literal['png', 'jpg', 'webp']


class LangDict(TypedDict, total=False):
    """A dictionary representing the different languages supported by xivapi."""

    en: str
    de: str
    fr: str
    ja: str



"""Tests related to xivapy models."""

from typing_extensions import Annotated
from xivapy.model import Model, FieldMapping


def test_model_sheet_name_from_class():
    """Test that we get the class name when no sheet name is defined."""

    class TestItem(Model):
        name: str

    assert TestItem.get_sheet_name() == 'TestItem'


def test_model_sheet_name_from_sheetname():
    """Test getting sheet name from defined field."""

    class CustomModel(Model):
        name: str
        __sheetname__ = 'Ackchyally'

    assert CustomModel.get_sheet_name() == 'Ackchyally'


def test_basic_model_validation():
    """Test that fields map even without specific fields defined."""

    class SimpleModel(Model):
        row_id: int
        name: str
        level: int = 0

    data = {'row_id': 1, 'name': 'Test', 'level': 50}
    model = SimpleModel.model_validate(data)

    assert model.row_id == 1
    assert model.name == 'Test'
    assert model.level == 50


def test_get_xivapi_fields_basic():
    """Test that defined basic fields in model generate as Title Cased names for xivapi."""

    class BasicModel(Model):
        row_id: int
        Name: str
        level: int

    fields = BasicModel.get_xivapi_fields()
    expected = {'row_id', 'Name', 'level'}
    assert fields == expected


def test_get_xivapi_fields_with_override():
    """Test overriding a field name to an xivapi-specific alias."""

    class BasicModel(Model):
        id: Annotated[int, FieldMapping('row_id')]
        name: str

    fields = BasicModel.get_xivapi_fields()
    expected = {'row_id', 'name'}
    assert fields == expected


def test_multiple_fields_with_same_nested_source():
    """Test that data is non-destructively pulled out of a nested dict."""

    class TestModel(Model):
        tanks: Annotated[int, FieldMapping('ContentMemberType.TanksPerParty')]
        healers: Annotated[int, FieldMapping('ContentMemberType.HealersPerParty')]

    data = {
        'ContentMemberType': {
            'fields': {
                'TanksPerParty': 2,
                'HealersPerParty': 1,
            }
        }
    }

    result = TestModel.model_validate(data)
    assert result.tanks == 2
    assert result.healers == 1


def test_model_with_no_fields():
    """Test defining an empty model."""

    class EmptyModel(Model): ...

    fields = EmptyModel.get_xivapi_fields()
    assert fields == set()



"""Tests related to xivapy.Client."""

from typing import Annotated
from pytest_httpx import HTTPXMock
import httpx
import pytest

from xivapy.client import Client, SearchResult
from xivapy.model import Model
from xivapy.exceptions import XIVAPIHTTPError

from tests.fixtures.api_responses import (
    BASIC_SEARCH_RESPONSE,
    SEARCH_RESPONSE_PAGE_1,
    SEARCH_RESPONSE_PAGE_2,
    VERSIONS_RESPONSE,
    SHEETS_RESPONSE,
    SHEET_ROW_RESPONSE,
)
from xivapy.model import FieldMapping


async def test_client_close():
    """Test that the client closes without exception."""
    client = Client()
    # No exception is essentially good
    await client.close()


def test_setting_patch():
    """Test setting patch as part of the client."""
    client = Client()
    client.patch('7.21')
    assert client.game_version == '7.21'


def test_flatten_item_data():
    """Test flattening api response data."""
    client = Client()
    data = {'row_id': 123, 'fields': {'Name': 'Foo'}}
    result = client._flatten_item_data(data)
    assert result == {'Name': 'Foo', 'row_id': 123}


async def test_versions_success(httpx_mock: HTTPXMock):
    """Test version endpoint with good response."""
    httpx_mock.add_response(
        url='https://v2.xivapi.com/api/version',
        json=VERSIONS_RESPONSE,
    )

    async with Client() as client:
        versions = await client.versions()
        assert '7.3x1' in versions
        assert 'latest' in versions


async def test_versions_http_error(httpx_mock: HTTPXMock):
    """Test version endpoint with bad response."""
    httpx_mock.add_response(
        url='https://v2.xivapi.com/api/version',
        status_code=500,
    )

    async with Client() as client:
        with pytest.raises(XIVAPIHTTPError) as exc_info:
            await client.versions()
        assert exc_info.value.status_code == 500


async def test_sheets_success(httpx_mock: HTTPXMock):
    """Test sheets endpoint with good response."""
    httpx_mock.add_response(
        url='https://v2.xivapi.com/api/sheet?version=latest',
        json=SHEETS_RESPONSE,
    )

    async with Client() as client:
        sheets = await client.sheets()
        assert 'Item' in sheets
        assert 'ContentFinderCondition' in sheets
        assert 'Quest' in sheets


async def test_sheets_http_error(httpx_mock: HTTPXMock):
    """Test sheets endpoint with bad response."""
    httpx_mock.add_response(
        url='https://v2.xivapi.com/api/sheet?version=latest',
        status_code=500,
    )

    async with Client() as client:
        with pytest.raises(XIVAPIHTTPError) as exc_info:
            await client.sheets()
        assert exc_info.value.status_code == 500


async def test_map_success(httpx_mock: HTTPXMock):
    """Test map endpoint with valid territory and index format."""
    httpx_mock.add_response(
        url='https://v2.xivapi.com/api/asset/map/a1b2/00?version=latest',
        content=b'abadlydrawnmapwithcrayonthatsnotevenajpg',
    )

    async with Client() as client:
        looking_for_a_good_map = await client.map('a1b2', '00')
        assert looking_for_a_good_map == b'abadlydrawnmapwithcrayonthatsnotevenajpg'


async def test_map_invalid_territory():
    """Test map with invalid territory format."""
    async with Client() as client:
        with pytest.raises(ValueError, match='Territory must be 4 characters'):
            await client.map('invalid', '00')


async def test_map_invalid_index():
    """Test map with invalid index."""
    async with Client() as client:
        with pytest.raises(
            ValueError, match='Index must be a 2-digit zero-padded number'
        ):
            await client.map('a1b2', 'invalid')


async def test_asset_success(httpx_mock: HTTPXMock):
    """Test asset with good response."""
    httpx_mock.add_response(
        url='https://v2.xivapi.com/api/asset?path=ui/icon/ultima.tex&format=png&version=latest',
        content=b'asparklerthathealstheenemy',
    )

    async with Client() as client:
        final_spell_icon = await client.asset(path='ui/icon/ultima.tex', format='png')
        assert final_spell_icon == b'asparklerthathealstheenemy'


async def test_asset_http_error(httpx_mock: HTTPXMock):
    """Test asset endpoint with bad response."""
    httpx_mock.add_response(
        url='https://v2.xivapi.com/api/asset?path=ui/icon/solution.tex&format=png&version=latest',
        status_code=500,
    )

    async with Client() as client:
        with pytest.raises(XIVAPIHTTPError, match='Failed to get asset') as exc_info:
            await client.asset(path='ui/icon/solution.tex', format='png')
        assert exc_info.value.status_code == 500


async def test_asset_none_found(httpx_mock: HTTPXMock):
    """Test asset endpoint where it isn't found."""
    httpx_mock.add_response(
        url='https://v2.xivapi.com/api/asset?path=ui/icon/selene.tex&format=png&version=latest',
        status_code=404,
    )

    async with Client() as client:
        asset = await client.asset(path='ui/icon/selene.tex', format='png')
        assert asset == None


async def test_search_success(httpx_mock: HTTPXMock):
    """Test searching something where it's a single result."""

    class TestSheet(Model):
        id: Annotated[int, FieldMapping('row_id')]
        name: Annotated[str, FieldMapping('Name')]
        level: Annotated[int, FieldMapping('Level')]

    expected_fields = TestSheet.get_fields_str()
    httpx_mock.add_response(
        url=httpx.URL(
            'https://v2.xivapi.com/api/search',
            params={
                'sheets': 'TestSheet',
                'query': '+Name="Test Item" +Level=50',
                'fields': expected_fields,
                'version': 'latest',
            },
        ),
        json=BASIC_SEARCH_RESPONSE,
    )

    client = Client()

    res_iter = aiter(client.search(TestSheet, query='+Name="Test Item" +Level=50'))

    item = await anext(res_iter)
    assert isinstance(item, SearchResult)
    assert item.score == pytest.approx(1.0)
    assert item.sheet == 'TestSheet'
    assert isinstance(item.data, TestSheet)
    assert item.data.id == 1
    assert item.data.name == 'Test Item'
    assert item.data.level == 50


# @pytest.mark.httpx_mock(assert_all_responses_were_requested=False)
async def test_paginated_search_success(httpx_mock: HTTPXMock):
    """Test getting multiple pages from the search endpoint with a cursor."""

    class TestSheet(Model):
        id: Annotated[int, FieldMapping('row_id')]
        name: Annotated[str, FieldMapping('Name')]
        level: Annotated[int, FieldMapping('Level')]

    expected_fields = TestSheet.get_fields_str()

    httpx_mock.add_response(
        url=httpx.URL(
            'https://v2.xivapi.com/api/search',
            params={
                'sheets': 'TestSheet',
                'query': 'Name~"Test Item" Level=50',
                'fields': expected_fields,
                'version': 'latest',
            },
        ),
        json=SEARCH_RESPONSE_PAGE_1,
    )
    httpx_mock.add_response(
        url=httpx.URL(
            'https://v2.xivapi.com/api/search',
            params={
                'sheets': 'TestSheet',
                'cursor': '28433b5b-7860-4395-88df-17c75c173a7c',
                'fields': expected_fields,
                'version': 'latest',
            },
        ),
        json=SEARCH_RESPONSE_PAGE_2,
    )

    client = Client()

    req_iter = aiter(client.search(TestSheet, query='Name~"Test Item" Level=50'))

    item = await anext(req_iter)
    assert isinstance(item, SearchResult)
    assert item.data.name == 'Test Item'
    assert item.data.level == 89

    item = await anext(req_iter)
    assert isinstance(item, SearchResult)
    assert item.data.name == 'Another Test Item'
    assert item.data.level == 50

    with pytest.raises(StopAsyncIteration):
        await anext(req_iter)



"""Classes related to xivapi client functions (fetching sheets, searching, etc)."""

from __future__ import annotations

from typing import AsyncIterable, Any, Self, Coroutine, cast, Sequence
from collections.abc import Iterable
from typing import Optional, AsyncIterator, overload
from itertools import batched
from dataclasses import dataclass
from re import match

import httpx
from aiostream.stream import chunks
from pydantic import ValidationError

from xivapy.model import Model
from xivapy.query import QueryBuilder
from xivapy.types import Format
from xivapy.exceptions import XIVAPIHTTPError, ModelValidationError
from xivapy.version import VERSION

__all__ = ['Client', 'SearchResult']


@dataclass
class SearchResult[T]:
    """Returned as part of using `xivapy.Client`'s search() method."""

    score: float
    sheet: str
    row_id: int
    data: T


class Client:
    """Async client for gathering data from xivapi rest api.

    Provides methods that touch all documented endpoints in xivapi. For information on
    those endpoints, please read https://v2.xivapi.com/api/docs.

    Args:
        base_url: Base URL for xivapi host; defaults to 'https://v2.xivapi.com'
        base_api_path: API path prevfix; defaults to '/api'
        game_version: Default game version for requests; defaults to 'latest'
        schema_version: Default schema version to use for requests
        batch_size: For the sheets endpoint, it will fetch in batches of that size

    Example:
        ```python
        class Item(xivapy.Model):
            id: Annotated[int, xivapy.FieldMapping('row_id')]
        client = xivapy.Client()
        item = await client.sheet(Item, row=123)
        print(item)
        await client.close()
        ```
    """

    def __init__(
        self,
        base_url: str = 'https://v2.xivapi.com',
        base_api_path: str = '/api',
        game_version: str = 'latest',
        schema_version: Optional[str] = None,
        batch_size: int = 100,
    ) -> None:
        """Initialize the Client with the given parameters."""
        self.base_url = base_url
        self.base_api_path = base_api_path
        transport = httpx.AsyncHTTPTransport(retries=3)
        self._client = httpx.AsyncClient(
            base_url=base_url,
            timeout=30.0,
            transport=transport,
            # TODO: let people set their own UA for this
            headers={'User-Agent': f'xivapi/{VERSION}'},
            limits=httpx.Limits(max_keepalive_connections=5, max_connections=10),
        )
        self.game_version = game_version
        self.schema_version = schema_version
        self.batch_size = batch_size

    async def close(self) -> None:
        """Close the interior HTTP client."""
        await self._client.aclose()

    # TODO: is there a better way to do this?
    def _add_version_params(self, params: dict) -> None:
        """Add version and schema parameters to request params if not already present."""
        if 'version' not in params:
            params['version'] = self.game_version
        if 'schema' not in params and self.schema_version:
            params['schema'] = self.schema_version

    def _flatten_item_data(self, data: dict) -> dict:
        """Extract and flatten row data from API response."""
        if not data or 'row_id' not in data:
            # TODO: maybe raise an exception or something?
            # Returning {} feels like losing data
            return {}

        processed_data = data.get('fields', {})
        processed_data['row_id'] = data['row_id']

        return processed_data

    async def __aenter__(self) -> Self:
        """Begin an async context where the client closes itself afterwards."""
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        """Exit the async context and close the session."""
        await self._client.aclose()

    def patch(self, version: str) -> None:
        """Sets the version for all endpoints to the version provided."""
        self.game_version = version

    async def versions(self) -> list[str]:
        """Retrieve a list of available game versions supported by the API."""
        try:
            response = await self._client.get(f'{self.base_api_path}/version')
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            raise XIVAPIHTTPError(
                f'Failed to get versions: {e}',
                status_code=e.response.status_code,
                response=e.response,
            )

        data = response.json()

        # flatten data
        version_names = []
        for version in data.get('versions', []):
            version_names.extend(version.get('names', []))

        return version_names

    # TODO: I wonder if index could just be a number that we 0-pad
    async def map(
        self, territory: str, index: str, version: Optional[str] = None
    ) -> Optional[bytes]:
        """Retrieve a composed map from the api.

        Args:
            territory: A 4-character string in the format of [letter][number][letter][number]
            index: A 2-digit (0-padded) numerical string
            version: A game version to fetch from

        Returns:
            A bytes object that that is a jpeg formatted map, or None
        """
        if not match(r'^[a-zA-Z]\d[a-zA-Z]\d$', territory):
            raise ValueError(
                f'Territory must be 4 characters in format [letter][digit][letter][digit], got: {territory}'
            )
        if not match(r'\d{2}$', index):
            raise ValueError(f'Index must be a 2-digit zero-padded number, got {index}')

        params = {}
        if version is not None:
            params['version'] = version
        self._add_version_params(params)

        try:
            response = await self._client.get(
                f'{self.base_api_path}/asset/map/{territory}/{index}', params=params
            )
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 404:
                return None
            raise XIVAPIHTTPError(
                f'Failed to get map {territory}/{index}: {e}',
                status_code=e.response.status_code,
                response=e.response,
            )

        return response.content

    async def sheets(self, version: Optional[str] = None) -> list[str]:
        """Gets a list of all sheets supported by the api."""
        params = {}
        if version is not None:
            params['version'] = version
        self._add_version_params(params)
        try:
            response = await self._client.get(
                f'{self.base_api_path}/sheet', params=params
            )
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            raise XIVAPIHTTPError(
                f'Failed to get sheets: {e}',
                status_code=e.response.status_code,
                response=e.response,
            )

        data = response.json()

        # massage data
        sheets = []
        for sheet in data.get('sheets', []):
            if (name := sheet.get('name')) is not None:
                sheets.append(name)

        return sheets

    @overload
    def search[T: Model](
        self,
        model_spec: type[T],
        query: QueryBuilder | str,
        **params,
    ) -> AsyncIterator[SearchResult[T]]: ...
    @overload
    def search[T1: Model, T2: Model](
        self,
        model_spec: tuple[type[T1], type[T2]],
        query: QueryBuilder | str,
        **params,
    ) -> AsyncIterator[SearchResult[T1 | T2]]: ...
    @overload
    def search[T1: Model, T2: Model, T3: Model](
        self,
        model_spec: tuple[type[T1], type[T2], type[T3]],
        query: QueryBuilder | str,
        **params,
    ) -> AsyncIterator[SearchResult[T1 | T2 | T3]]: ...
    def search(
        self,
        model_spec: type[Model] | tuple[type[Model], ...],
        query: QueryBuilder | str,
        **params,
    ) -> Any:
        """Search XIVAPI for data using a query.

        Args:
            model_spec: Model class or tuple of model classes to search the sheets for.
            query: A QueryBuilder search or a plain string with the search terms
            **params: Additional search parameters

        Returns:
            AsyncIterator yielding SearchResult objects with data typed against your model_spec

        Example:
            ```python
            class ContentFinderCondition(Model):
                name: Annotated[str, FieldMapping('Name')]
                content_type: Annotated[str, FieldMapping('ContentType.Name')]

            async for result in client.search(ContentFinderCondition, 'ContentType.Name="Trials"'):
                print(f'{result.data.name} - {result.data.content_type}')

            query = QueryBuilder().where(ItemLevelRequired=690).contains(Name='extreme')
            async for result in client.search(ContentFinderCondition, query):
                # result.data is still properly typed
            ```
        """
        return self._search_impl(model_spec, query, **params)

    async def _search_impl(
        self,
        model_spec: type[Model] | tuple[type[Model], ...],
        query: QueryBuilder | str,
        **params,
    ) -> AsyncIterator[SearchResult[Model]]:
        """The underlying search implementation method."""
        # EVERYTHING MUST BE TUPLES
        models: tuple[type[Model], ...]
        if isinstance(model_spec, type):
            models = (model_spec,)
        else:
            models = model_spec

        sheets = {model.get_sheet_name() for model in models}

        if 'fields' not in params:
            fields = {model.get_fields_str() for model in models}
            params['fields'] = ','.join(fields)

        if isinstance(query, QueryBuilder):
            query_str = query.build()
        else:
            query_str = str(query)

        search_params = {'sheets': ','.join(sheets), 'query': query_str, **params}
        self._add_version_params(search_params)

        # Create model lookup table
        model_lut = {model.get_sheet_name(): model for model in models}

        cursor = None

        while True:
            current_params = search_params.copy()
            if cursor:
                current_params['cursor'] = cursor
                current_params.pop('query', None)

            try:
                response = await self._client.get(
                    f'{self.base_api_path}/search', params=current_params
                )
                response.raise_for_status()
            except httpx.HTTPStatusError as e:
                raise XIVAPIHTTPError(
                    f'Search failed: {e}',
                    status_code=e.response.status_code,
                    response=e.response,
                )

            data = response.json()

            for result in data.get('results', []):
                sheet_name = result.get('sheet')
                if sheet_name in model_lut:
                    model_class = model_lut[sheet_name]

                    processed_data = self._flatten_item_data(
                        {
                            'row_id': result['row_id'],
                            'fields': result.get('fields', {}),
                        }
                    )

                    try:
                        model_instance = model_class.model_validate(processed_data)
                    except ValidationError as e:
                        raise ModelValidationError(model_class, e, processed_data)
                    yield SearchResult(
                        score=result.get('score', 0.0),
                        sheet=sheet_name,
                        row_id=result['row_id'],
                        data=model_instance,
                    )
            # Are there more pages?
            cursor = data.get('next')
            if not cursor:
                break

    async def asset(
        self, path: str, format: Format = 'png', version: Optional[str] = None
    ) -> Optional[bytes]:
        """Fetches an asset from the game, formatted into a more user-friendly format.

        Args:
            path: The path to fetch from
            format: The format for the resource to be formatted as ('jpg', 'png', 'webp')
            version: A game version to fetch from

        Returns:
            A bytes object in the format requested, or None if not found.
        """
        params = {
            'path': path,
            'format': format,
        }
        if version is not None:
            params['version'] = version
        self._add_version_params(params)
        try:
            response = await self._client.get(
                f'{self.base_api_path}/asset', params=params
            )
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 404:
                return None
            raise XIVAPIHTTPError(
                f'Failed to get asset {path}: {e}',
                status_code=e.response.status_code,
                response=e.response,
            )

        return response.content

    async def icon(
        self, icon_id: int, format: Format = 'jpg', version: Optional[str] = None
    ) -> Optional[bytes]:
        """Fetches an icon resource from the game by id.

        Args:
            icon_id: The icon id in game data.
            format: The format for the icon to be in (defaults to 'jpg')
            version: A game version to fetch from

        Returns:
            A bytes object of the icon in the selected format, or None if not found.
        """
        folder = f'{icon_id // 1000 * 1000:06d}'
        path = f'ui/icon/{folder}/{icon_id:06d}_hr1.tex'
        return await self.asset(path, format=format, version=version)

    @overload
    def sheet[T: Model](
        self,
        model_class: type[T],
        *,
        row: int,
        **params,
    ) -> Coroutine[Any, Any, Optional[T]]: ...
    @overload
    def sheet[T: Model](
        self,
        model_class: type[T],
        *,
        rows: Iterable[int] | AsyncIterable[int],
        **params,
    ) -> AsyncIterator[T]: ...
    def sheet[T: Model](
        self,
        model_class: type[T],
        *,
        row: Optional[int] = None,
        rows: Optional[Iterable[int] | AsyncIterable[int]] = None,
        **params,
    ) -> Coroutine[Any, Any, Optional[T]] | AsyncIterator[T]:
        """Fetch one or more rows from a sheet.

        Args:
            model_class: An xivapy.Model class for the results to be coerced to
            row: A single row id to fetch
            rows: Multiple row ids to fetch
            **params: Extra parameters which are passed to the sheets endpoint

        Returns:
            This returns either:
            * A single row formatted by the model given
            * An AsyncIterator of the rows requested
            * None in the case of the item not being found
        """
        if row is not None and rows is not None:
            raise ValueError("Cannot specify both 'row' and 'rows'")

        if row is not None:
            return self._get_single_row(model_class, row, **params)
        elif rows is not None:
            return self._get_multiple_rows(model_class, rows, **params)
        else:
            raise ValueError("Must specify either 'row' or 'rows'")

    async def _get_single_row[T: Model](
        self,
        model_class: type[T],
        row: int,
        **params,
    ) -> Optional[T]:
        """An internal method for fetching a single row."""
        self._add_version_params(params)
        if 'fields' not in params:
            params['fields'] = model_class.get_fields_str()

        try:
            response = await self._client.get(
                f'{self.base_api_path}/sheet/{model_class.get_sheet_name()}/{row}',
                params=params,
            )
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 404:
                return None
            raise XIVAPIHTTPError(
                f'Failed to get sheet rows for {model_class.get_sheet_name()}: {e}',
                status_code=e.response.status_code,
                response=e.response,
            )

        data = response.json()
        if not data or 'row_id' not in data:
            return None
        processed_data = self._flatten_item_data(data)

        try:
            return model_class.model_validate(processed_data)
        except ValidationError as e:
            raise ModelValidationError(model_class, e, processed_data)

    async def _get_multiple_rows[T: Model](
        self,
        model_class: type[T],
        rows: Iterable[int] | AsyncIterable[int],
        **params,
    ) -> AsyncIterator[T]:
        """An internal method for fetching multiple rows."""
        self._add_version_params(params)
        if 'fields' not in params:
            params['fields'] = model_class.get_fields_str()

        if hasattr(rows, '__aiter__'):
            # mypy can't resolve aiostream types correctly in some cases - see upstream issue:
            # https://github.com/vxgmichel/aiostream/issues/105
            async with chunks(rows, self.batch_size).stream() as streamer:  # type: ignore[arg-type,var-annotated]
                async for batch in streamer:
                    batch_seq = cast(Sequence[int], batch)
                    async for item in self._process_batch(
                        model_class, batch_seq, **params
                    ):
                        yield item
        else:
            for batch in batched(rows, self.batch_size):  # type: ignore[assignment]
                async for item in self._process_batch(model_class, batch, **params):
                    yield item

    async def _process_batch[T: Model](
        self, model_class: type[T], batch: Sequence[int], **params
    ) -> AsyncIterator[T]:
        # TODO: allow overriding batch-size in sheet
        rows_param = ','.join(str(id) for id in batch)

        try:
            response = await self._client.get(
                f'{self.base_api_path}/sheet/{model_class.get_sheet_name()}',
                params={**params, 'rows': rows_param},
            )
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            raise XIVAPIHTTPError(
                '', status_code=e.response.status_code, response=e.response
            )

        data = response.json()

        for item_data in data.get('rows', []):
            if not item_data or 'row_id' not in item_data:
                continue

            processed_data = self._flatten_item_data(item_data)
            try:
                yield model_class.model_validate(processed_data)
            except ValidationError as e:
                raise ModelValidationError(model_class, e, processed_data)



"""Exceptions for xivapy."""

from typing import Optional, Any

import httpx
from pydantic import ValidationError

__all__ = [
    'XIVAPIError',
    'XIVAPIHTTPError',
    'ModelValidationError',
    'QueryBuildError',
]


class XIVAPIError(Exception):
    """Base exception for all xivapy-related errors."""

    def __init__(self, message: str, details: Optional[dict[str, Any]] = None) -> None:
        """Base exception for all xivapy-related errors.

        This exception will be raised if none of the other exceptions are appropriate;
        otherwise this acts as a base class for all xivapy-related exceptions.

        Args:
            message: Standard Exception message
            details: An optional dict[str, Any] with information about the exception
        """
        super().__init__(message)
        self.details = details or {}


class XIVAPIHTTPError(XIVAPIError):
    """HTTP-related errors when communicating with XIVAPI."""

    def __init__(
        self, message: str, status_code: int, response: Optional[httpx.Response] = None
    ) -> None:
        """Raised when the upstream service raises an issue.

        Typically, the library will try to recover unless the api has provided
        data in a format we don't understand (because it was changed), or the
        server indicates an issue with an obvious status code of failure (e.g., 500)

        Args:
            message: Standard Exception message
            status_code: The return code from upstream
            response: Optional upstream httpx.Response object, if any
        """
        super().__init__(message)
        self.status_code = status_code
        self.response = response


class ModelValidationError(XIVAPIError):
    """Raised when API response data doesn't validate against the model schema."""

    def __init__(
        self,
        model_class: type,
        validation_error: ValidationError,
        raw_data: Optional[dict] = None,
    ) -> None:
        """Raised when xivapy.Model fails to validate.

        Essentially a wrapper around pydantic's ValidationError, but
        also provides the raw data that failed to validate for the
        consumer to perhaps figure out what went wrong.

        Args:
            model_class: The model that failed validation
            validation_error: A pydantic validation error explaining what failed to validate
            raw_data: The data that failed to validate
        """
        message = f'Failed to validate data for model {model_class.__name__}: {validation_error}'
        super().__init__(message)
        self.model_class = model_class
        self.validation_error = validation_error
        self.raw_data = raw_data


class QueryBuildError(XIVAPIError):
    """Raised when there's an error building a query."""

    def __init__(self, message: str, query_parts: Optional[list[str]] = None) -> None:
        """Raised when queries fail to build.

        This could be caused in validation situations (such as marking a
        query as both required and excluded), or in situations where the
        query will fail to build for some exotic reason.

        Args:
            message: Standard Exception message
            query_parts: The parts of the query that failed
        """
        super().__init__(message)
        self.query_parts = query_parts or []



"""Query building utilities for xivapi search."""

from __future__ import annotations

from typing import Self, Any
from dataclasses import dataclass

from xivapy.exceptions import QueryBuildError

__all__ = [
    'Query',
    'QueryBuilder',
    'Group',
]


@dataclass
class Query:
    """Represents a composable query unit in xivapi's query interface.

    Provides an abstracted interface for dealing with xivapi-style
    query elements, which are created automatically as part of QueryBuilder's
    methods.

    Example:
        >>> query = (QueryBuilder()
        ...    .custom('Name', '=', 'Alexander', required=True)
        ...    .custom('ItemLevel', '>=', 50)
        ...    .required())
        >>> print(query.build())
        +Name="Alexander" +ItemLevel>=50
    """

    field: str
    operation: str
    # TODO: scope this to number | str | bool as those are the only
    # listed ones in the docs
    value: Any
    required: bool = False
    excluded: bool = False

    def __str__(self) -> str:
        """Returns a string representation of the query."""
        if self.required and self.excluded:
            raise QueryBuildError('Query cannot be set to both required and excluded')
        if self.required:
            prefix = '+'
        elif self.excluded:
            prefix = '-'
        else:
            prefix = ''
        if isinstance(self.value, str) and self.operation in ['=', '~']:
            escaped_value = f'"{self.value}"'
        elif isinstance(self.value, bool):
            escaped_value = 'true' if self.value else 'false'
        else:
            escaped_value = str(self.value)
        return f'{prefix}{self.field}{self.operation}{escaped_value}'


class QueryBuilder:
    """Builder for constructing xivapi search queries.

    Provides an abstracted interface for dealing with xivapi-style
    search queries, including grouping, operators like >=, ~, =, etc.
    Queries can be chained and grouped to create more complex queries.

    Example:
        >>> query = (QueryBuilder()
        ...    .contains(Name='Alexander')
        ...    .gte(ItemLevel=50)
        ...    .required())
        >>> print(query.build())
        Name="Alexander" +ItemLevel>=50
    """

    def __init__(self) -> None:
        """Initializes an empty query builder."""
        self.clauses: list[Query | Group] = []

    def where(self, **kwargs) -> Self:
        """Add an equality condition to the query.

        Args:
            **kwargs: Field-value pairs for exact matches.

        Returns:
            Self for method chaining.
        """
        for field, value in kwargs.items():
            self.clauses.append(Query(field, '=', value))
        return self

    def contains(self, **kwargs) -> Self:
        """Add a partial string match to the query.

        Args:
            **kwargs: Field-value pairs for partial matching.

        Returns:
            Self for method chaining.
        """
        for field, value in kwargs.items():
            self.clauses.append(Query(field, '~', value))
        return self

    def gt(self, **kwargs) -> Self:
        """Add a greater than (>) numeric comparison to the query.

        Args:
            **kwargs: Field-value pairs for comparison.

        Returns:
            Self for method chaining.
        """
        for field, value in kwargs.items():
            self.clauses.append(Query(field, '>', value))
        return self

    def gte(self, **kwargs) -> Self:
        """Add a greater than or equal (>=) numeric comparison to the query.

        Args:
            **kwargs: Field-value pairs for comparison.

        Returns:
            Self for method chaining.
        """
        for field, value in kwargs.items():
            self.clauses.append(Query(field, '>=', value))
        return self

    def lt(self, **kwargs) -> Self:
        """Add a less than (<) numeric comparison to the query.

        Args:
            **kwargs: Field-value pairs for comparison.

        Returns:
            Self for method chaining.
        """
        for field, value in kwargs.items():
            self.clauses.append(Query(field, '<', value))
        return self

    def lte(self, **kwargs) -> Self:
        """Add a less than or equal (<=) numeric comparison to the query.

        Args:
            **kwargs: Field-value pairs for comparison.

        Returns:
            Self for method chaining.
        """
        for field, value in kwargs.items():
            self.clauses.append(Query(field, '<=', value))
        return self

    def required(self) -> Self:
        """Marks the previous query item as required.

        Returns:
            Self for method chaining.
        """
        if self.clauses:
            last = self.clauses[-1]
            if isinstance(last, Query):
                last.required = True
            elif isinstance(last, Group):
                last.required = True
        return self

    def excluded(self) -> Self:
        """Marks the previous query item as excluded.

        Returns:
            Self for method chaining.
        """
        if self.clauses:
            last = self.clauses[-1]
            if isinstance(last, Query):
                last.excluded = True
            elif isinstance(last, Group):
                last.excluded = True
        return self

    def custom(self, *items: Query) -> Self:
        """Allows injecting regular Queries into QueryBuilder.

        Args:
            items: one or more Query items to be added to the list of clauses to evaluate.

        Returns:
            Self for method chaining.
        """
        self.clauses.extend(items)
        return self

    def or_any(self, *items: Query | QueryBuilder) -> Self:
        """Creates a group to match agains.

        Args:
            items: one or more Query or QueryBuilder instances to add to the group.

        Returns:
            Self for method chaining.
        """
        self.clauses.append(Group(list(items)))
        return self

    def build(self) -> str:
        """Builds the current query as a string.

        Returns:
            A string representation of the query.
        """
        if not self.clauses:
            return ''
        return ' '.join(str(clause) for clause in self.clauses)

    def __str__(self) -> str:
        """Shows the current query as a string representation.

        Returns:
            A string representation of the query.
        """
        return self.build()


@dataclass
class Group:
    """Constructs groups to be used with QueryBuilder.

    Provides an abstracted interfaces around xivapi-style nested
    grouped query items.

    Example:
        >>> query = (QueryBuilder()
        ...    .or_any( # This creates a Group instance
        ...        QueryBuilder().where(Name="The Binding Coils of Bahamut"),
        ...        Query('ClassJob.Abbreviation', '=', 'PCT'))
        ...    .required())
        >>> print(query.build())
        +(Name="The Binding Coils of Bahamud" ClassJob.Abbreviation="PCT")
    """

    items: list[Query | QueryBuilder]
    required: bool = False
    excluded: bool = False

    def __str__(self) -> str:
        """Returns a string representation of the grouping."""
        if self.required and self.excluded:
            raise QueryBuildError('Query cannot be set to both required and excluded')
        if self.required:
            prefix = '+'
        elif self.excluded:
            prefix = '-'
        else:
            prefix = ''
        inner = ' '.join(str(item) for item in self.items)
        return f'{prefix}({inner})'
